"""

Manage the Websocket 'opcode' enumerations.

Not using enum34 purely to keep dependencies to a minimum.


"""

from __future__ import print_function
from __future__ import unicode_literals


class Opcode(object):
    """Enumeration of websocket opcodes."""

    CONTINUATION = 0
    TEXT = 1
    BINARY = 2
    RESERVED1 = 3
    RESERVED2 = 4
    RESERVED3 = 5
    RESERVED4 = 6
    RESERVED5 = 7
    CLOSE = 8
    PING = 9
    PONG = 0xA
    RESERVED6 = 0xB
    RESERVED7 = 0xC
    RESERVED8 = 0xD
    RESERVED9 = 0xE
    RESERVED10 = 0xF

    @classmethod
    def to_str(cls, opcode):
        return cls._opcode_to_str.get(opcode, '?')

Opcode._opcode_to_str = {
    getattr(Opcode, _name): _name
    for _name in dir(Opcode)
    if not _name.startswith('_') and _name.isupper()
}

reserved_opcodes = {
    Opcode.RESERVED1,
    Opcode.RESERVED2,
    Opcode.RESERVED3,
    Opcode.RESERVED4,
    Opcode.RESERVED5,
    Opcode.RESERVED6,
    Opcode.RESERVED7,
    Opcode.RESERVED8,
    Opcode.RESERVED9,
    Opcode.RESERVED10,
}


def is_reserved(opcode):
    """Check if an opcode is reserved."""
    return opcode in reserved_opcodes



"""
Coroutine Parser
================

This little but of 'magic' makes parsing a stream of bytes from a socket
a relatively pain-free exercise.

"""

from __future__ import unicode_literals


class ParseError(Exception):
    """Stream failed to parse."""


class _Awaitable(object):
    """An operation that effectively suspends the coroutine."""
    # Analogous to Python3 asyncio concept

    def validate(self, chunk):
        """Raise any ParseErrors"""


class _ReadBytes(_Awaitable):
    """Reads a fixed number of bytes."""
    __slots__ = ['remaining']
    def __init__(self, count):
        self.remaining = count


class _ReadUtf8(_ReadBytes):
    """Reads a fixed number of bytes, validates utf-8."""
    __slots__ = ['utf8_validator']
    def __init__(self, count, utf8_validator):
        self.remaining = count
        self.utf8_validator = utf8_validator

    def validate(self, data):
        valid, _, _, _ = self.utf8_validator.validate(data)
        if not valid:
            raise ParseError('invalid utf8')


class _ReadUntil(_Awaitable):
    """Read until a separator."""
    __slots__ = ['sep', 'max_bytes']
    def __init__(self, sep, max_bytes=None):
        self.sep = sep
        self.max_bytes = max_bytes

    def check_length(self, pos):
        """Check the length is within max bytes."""
        if self.max_bytes is not None and pos > self.max_bytes:
            raise ParseError(
                '{!r} expected in {} byte(s)',
                self.sep,
                self.max_bytes
            )


class Parser(object):
    """
    Coroutine based stream parser.

    Splits a steam of arbitrary sequences of bytes in to logical
    objects. The `feed` method will yield any results of parsing.

    Here's an example of use::

        while True:
            data = sock.recv(1024)
            for obj in parser.feed(data):
                self.on_obj(obj)

    The coroutine magic allows parsers to be non-blocking while still
    implemented in a simple procedural manner.

    """

    def __init__(self):
        self._gen = None
        self._awaiting = None
        self._buffer = []  # Buffer for reads
        self._until = b''  # Buffer for read untils
        self._closed = False
        self.reset()

    read = _ReadBytes
    read_utf8 = _ReadUtf8
    read_until = _ReadUntil

    def __del__(self):
        self.close()

    def reset(self):
        """Reset the parser, so it may be used on a fresh stream."""
        self._gen = self.parse()
        self._awaiting = next(self._gen)

    def close(self):
        """Close the parser."""
        if self._gen is not None:
            self._gen.close()
            self._gen = None

    def feed(self, data):
        """
        Called with data (bytes), will yield 0 or more objects parsed
        from the stream.

        :param bytes data: Data to decode.

        """

        def _check_length(pos):
            try:
                self._awaiting.check_length(pos)
            except ParseError as error:
                self._gen.throw(error)

        pos = 0
        while pos < len(data):
            # Awaiting a read of a fixed number of bytes
            if isinstance(self._awaiting, _ReadBytes):
                # This many bytes left to read
                remaining = self._awaiting.remaining
                # Bite off remaining bytes
                chunk = data[pos:pos + remaining]
                chunk_size = len(chunk)
                pos += chunk_size
                try:
                    # Validate new data
                    self._awaiting.validate(chunk)
                except ParseError as error:
                    # Raises an exception in parse()
                    self._gen.throw(error)
                # Add to buffer
                self._buffer.append(chunk)
                remaining -= chunk_size
                if remaining:
                    # Await more bytes
                    self._awaiting.remaining = remaining
                else:
                    # Got all the bytes we need in buffer
                    send_bytes = b''.join(self._buffer)
                    del self._buffer[:]
                    # Send to coroutine, get new 'awaitable'
                    self._awaiting = self._gen.send(send_bytes)

            # Awaiting a read until a terminator
            elif isinstance(self._awaiting, _ReadUntil):
                # Reading to separator
                chunk = data[pos:]
                self._until += chunk
                sep = self._awaiting.sep
                sep_index = self._until.find(sep)

                if sep_index == -1:
                    # Separator not found, advance position
                    pos += len(chunk)
                    _check_length(len(self._until))
                else:
                    # Found separator
                    # Get data prior to and including separator
                    _check_length(sep_index + len(sep))
                    split_index = sep_index + len(sep)
                    send_bytes = self._until[:split_index]
                    # Reset data, to continue parsing
                    data = self._until[split_index:]
                    self._until = b''
                    pos = 0
                    # Send bytes to coroutine, get new 'awaitable'
                    self._awaiting = self._gen.send(send_bytes)
            # Yield any non-awaitables...
            while not isinstance(self._awaiting, _Awaitable):
                yield self._awaiting
                self._awaiting = next(self._gen)

    def parse(self):
        """
        A generator to parse incoming stream.

        Yield the result of `self.read` to read n bytes from the stream.
        Yield any parsed objects

        Here's an example::

            size = struct.unpack('<I', yield self.read(4))
            data = yield self.read(size)
            yield data.decode()

        """
        yield


if __name__ == "__main__":  # pragma: no cover
    class TestParser(Parser):
        def parse(self):
            data = yield self.read_until(b'\r\n\r\n')
            yield data
            data = yield self.read(1)
            yield data
            data = yield self.read(2)
            yield data
            data = yield self.read(4)
            yield data
            data = yield self.read(2)
            yield data
    parser = TestParser()
    for b in (b'head', b'ers: example', b'\r\n', b'\r\n', b'12', b'34', b'5', b'678', b'', b'90'):
        for frame in parser.feed(b):
            print(repr(frame))



from tornado import gen, ioloop, web, websocket

INTRO = b"""\
  _                                     _
 | |                                   | |
 | |     ___  _ __ ___   ___  _ __   __| |
 | |    / _ \| '_ ` _ \ / _ \| '_ \ / _` |
 | |___| (_) | | | | | | (_) | | | | (_| |
 |______\___/|_| |_| |_|\___/|_| |_|\__,_|
   _____ _           _    _____
  / ____| |         | |  / ____|
 | |    | |__   __ _| |_| (___   ___ _ ____   _____ _ __
 | |    | '_ \ / _` | __|\___ \ / _ \ '__\ \ / / _ \ '__|
 | |____| | | | (_| | |_ ____) |  __/ |   \ V /  __/ |
  \_____|_| |_|\__,_|\__|_____/ \___|_|    \_/ \___|_|

 The simplest chat server imaginable.
"""

def col(text):
    return u'\x1b[1;34m{}\x1b[0m'.format(text)

class ChatHandler(websocket.WebSocketHandler):

    chatting = set()

    def check_origin(self, origin):
        return True

    @gen.coroutine
    def open(self):
        self.set_nodelay(True)
        self.chatting.add(self)
        yield self.write_message(col(INTRO))

    def on_close(self):
        self.chatting.remove(self)

    @gen.coroutine
    def on_message(self, chat):
        for chatter in self.chatting:
            if chatter is not self:
                yield chatter.write_message(col(chat))

app = web.Application([
    (r'^/chat/$', ChatHandler),
])
app.listen(9001)
ioloop.IOLoop.instance().start()